"""Command line interfaces."""
from __future__ import annotations
import importlib

__all__ = ["deploy", "service", "migrate", "server_map"]


def __getattr__(name):
    return getattr(importlib.import_module(f"._{name}", __name__), "cli")



#!/usr/bin/env python3
from __future__ import annotations
from collections import defaultdict
from pathlib import Path
import os
from typing import NamedTuple

from flask import Flask, request
from flask_restful import reqparse, Resource, Api
import toml

ServiceInfo = NamedTuple(
    "ServiceInfo", [("name", str), ("python_path", str), ("hosts", str)]
)


class ServerBaseClass:
    """Base class for looking/saving server information from/to disk."""

    def __init__(self):
        self.server_information: dict[str, list[ServiceInfo]] = defaultdict(
            list[ServiceInfo]
        )
        self.server_info_file.touch()
        self.update_server_information()

    @property
    def server_info_file(self) -> Path:
        """Path to the filename holding the server infromation."""
        return Path(os.environ["SERVER_FILE"])

    def update_server_information(self) -> None:
        """Read the server information."""
        with self.server_info_file.open() as f_obj:
            server_dict = toml.load(f_obj)
        for project, settings in server_dict.items():
            for service, (python_path, hosts) in settings.items():
                self.server_information[project].append(
                    ServiceInfo(name=service, python_path=python_path, hosts=hosts)
                )

    def _update(self) -> None:
        with self.server_info_file.open("w") as f_obj:
            toml.dump(self._server_info_tuple_to_dict(), f_obj)
        self.update_server_information()

    def _server_info_tuple_to_dict(self) -> dict[str, dict[str, tuple[str, str]]]:
        """Convert ServiceInfo namedtuple to dictionary."""
        dump_dict: dict[str, dict[str, tuple[str, str]]] = {}
        for project, settings in self.server_information.items():
            dump_dict[project] = {}
            for settings_tuple in settings:
                dump_dict[project][settings_tuple.name] = (
                    settings_tuple.python_path,
                    settings_tuple.hosts,
                )
        return dump_dict


class ServerLookup(ServerBaseClass, Resource):
    """Lookup server information."""

    def __init__(self):
        super().__init__()

    def get(self) -> tuple[dict[str, dict[str, tuple[str, str]]], int]:
        """Get method, for getting server information.

        Returns:
        --------
            dict[str dict[str, tuple[str, str]]]:  sever information
        """
        dump_dict = self._server_info_tuple_to_dict()
        return dump_dict, 200


class ServerEntry(ServerBaseClass, Resource):
    """Add entries to the server map."""

    def __init__(self):
        super().__init__()

    def put(self, project: str) -> tuple[str, int]:
        """Put method, for setting a server entry.

        Parameters
        -----------
        project: str
            The freva project name which is updated.

        Returns
        --------
            dict: status information
        """
        try:
            config = toml.loads(request.form["config"])
        except toml.TomlDecodeError as error:
            return error.__str__(), 400
        for service, settings in config.items():
            try:
                settings_tuple = ServiceInfo(name=service, **settings)
            except TypeError:
                return "TypeError: Wrong settings", 400
            self.server_information[project].append(settings_tuple)
        try:
            self._update()
        except Exception as e:
            return e.__str__(), 500
        return "updated", 201

    def get(self, project: str) -> tuple[dict[str, tuple[str, str]], int]:
        """Get method, for getting server information.

        Parameters
        -----------
        project: str
            The freva project name which is queried.

        Returns:
        --------
            dict[str, tuple[str, str]]:  sever information
        """

        try:
            return (
                self._server_info_tuple_to_dict()[project],
                200,
            )
        except KeyError:
            return f"No such project: {project}", 404


class ServerStaus(Resource):
    """Get and set the staus of services."""

    service_status: dict[str, dict[str, dict[str, str]]] = {}

    def get(self, project: str, service: str) -> tuple[dict[str, str], int]:
        """Get method for retreivng the service status.

        Parameters
        ----------
        project: str
            The freva project name which is updated.
        service: str
            Name of the service which status is retreived
        """
        try:
            return self.service_status[project][service], 200
        except KeyError:
            return f"No such service: {service}", 404

    def put(self, project: str, service: str) -> tuple[str, int]:
        """Put method, for setting the service status.

        Parameters
        -----------
        project: str
            The freva project name which is updated.

        Returns
        --------
            dict: operation status information
        """
        keys = ("mem", "status", "cpu")
        args = {k: v.lower() for (k, v) in request.form.items() if k in keys and v}
        if project not in self.service_status:
            self.service_status[project] = {service: args}
        else:
            for key, value in args.items():
                try:
                    self.service_status[project][service][key] = value
                except KeyError:
                    self.service_status[project][service] = {key: value}
        return "updated", 201


if __name__ == "__main__":
    app = Flask(__name__)
    api = Api(app)
    api.add_resource(ServerLookup, "/")
    api.add_resource(ServerEntry, "/<string:project>")
    api.add_resource(ServerStaus, "/<string:project>/<string:service>")
    app.run(host="0.0.0.0", port=5008)



"""Call the freva tui."""
from __future__ import annotations
import argparse

from freva_deployment import __version__
from freva_deployment.deploy import DeployFactory
from freva_deployment.utils import set_log_level, RichConsole
from .main_window import MainApp


def parse_args() -> int:
    """Construct command line argument parser."""
    app = argparse.ArgumentParser(
        prog="deploy-freva",
        description="Text User Interface for Freva deployment",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    app.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    app.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser = app.parse_args()
    return parser.verbose


def tui() -> None:
    """Create and run Text User Interface (TUI) for deployment."""
    verbosity = parse_args()
    set_log_level(verbosity)
    try:
        main_app = MainApp()
        main_app.run()
    except KeyboardInterrupt:
        try:
            main_app.thread_stop.set()
            main_app.save_config_to_file()
        except AttributeError:
            pass
        return
    setup = main_app.setup
    if setup:
        server_map = setup.pop("server_map")
        ask_pass = setup.pop("ask_pass")
        steps = ", ".join(setup["steps"])
        RichConsole.print(f"Playing steps: [i]{steps}[/] with ansible")
        with DeployFactory(**setup) as DF:
            DF.play(server_map, ask_pass, verbosity)



"""CLI to interact with running services."""
from __future__ import annotations
import argparse
from getpass import getuser
import json
import logging
from pathlib import Path
import shlex
from subprocess import run
import sys
from tempfile import NamedTemporaryFile

import appdirs
from numpy import sign

from freva_deployment import __version__
from ..utils import asset_dir, logger


def parse_args(args: list[str]) -> argparse.Namespace:
    """Consturct command line argument parser."""
    app = argparse.ArgumentParser(
        prog="deploy-freva-map",
        description="Create service that maps the freva server architecture.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    app.add_argument(
        "servername",
        type=str,
        help="The server name where the infrastructure mapping service is deployed",
    )
    app.add_argument(
        "--port",
        type=int,
        help="The port the service is listing to",
        default=6111,
    )
    app.add_argument(
        "--wipe", default=False, action="store_true", help="Delete any existing data."
    )
    app.add_argument(
        "--user",
        default=None,
        type=str,
        help="Username to log on to the target server.",
    )
    app.add_argument(
        "--python-path",
        default="/usr/bin/python",
        type=Path,
        help="Path to the default python3 interpreter on the target machine.",
    )
    app.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    app.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    argsp = app.parse_args(args)
    return argsp


YAML_INVENTORY = """
freva_map:
    hosts: {hosts}
    vars:
        ansible_python_interpreter: {python_path}
        port: {port}
        asset_dir: {asset_dir}
        wipe: {wipe}
"""


INPT_MSG = (
    "You should only deploy the server map once per domain/institution."
    " Are you sure you want to continue"
)


def cli(argv: list[str] | None = None) -> None:
    """Run command line interface."""
    argv = argv or sys.argv[1:]
    argp = parse_args(argv)
    user = argp.user or getuser()
    playbook = asset_dir / "playbooks" / "server-map-playbook.yml"
    logger.critical(INPT_MSG)
    logger.setLevel(max(logging.INFO - 10 * argp.verbose, logging.DEBUG))
    answer = input("Continue? [y|N]: ")
    verbosity = sign(argp.verbose) * "-" + argp.verbose * "v"
    if answer.lower() not in ("yes", "y"):
        logger.debug("Exiting routine.")
        return
    with NamedTemporaryFile() as inventory_file:
        inventory = YAML_INVENTORY.format(
            hosts=argp.servername,
            python_path=argp.python_path,
            port=argp.port,
            asset_dir=asset_dir,
            wipe=str(argp.wipe).lower(),
        )
        logger.debug(inventory)
        with open(inventory_file.name, "w") as f_obj:
            f_obj.write(inventory)
        cmd = shlex.split(
            f"ansible-playbook {verbosity} -u {user} --ask-pass --ask-become -i "
            f"{inventory_file.name} {playbook}"
        )
        run(cmd, check=True)
        cache_dir = Path(appdirs.user_cache_dir()) / "freva-deployment"
        host_name = f"{argp.servername}:{argp.port}"
        try:
            with (cache_dir / "freva_deployment.json").open("r") as f_obj:
                config = json.load(f_obj)
        except (FileNotFoundError, IOError, json.JSONDecodeError):
            config = {}
        config["server_map"] = host_name
        with (cache_dir / "freva_deployment.json").open("w") as f_obj:
            json.dump(config, f_obj)



from __future__ import annotations
import os
from pathlib import Path
import logging

import curses
import npyscreen


logging.basicConfig(level=logging.DEBUG)
logger: logging.Logger = logging.getLogger("deploy-freva-tui")


class FileSelector(npyscreen.FileSelector):
    """FileSelector widget that allows for filtering file extensions."""

    file_extentions: list[str]
    """List of allowed file extensions."""
    value: str
    """The value of this file selector."""

    def __init__(
        self, *args, file_extentions: str | list[str] = [".toml"], **kwargs
    ) -> None:

        if isinstance(file_extentions, str):
            self.file_extentions = [file_extentions]
        else:
            self.file_extentions = file_extentions
        super().__init__(*args, **kwargs)
        self.how_exited_handers[npyscreen.wgwidget.EXITED_ESCAPE] = self.exit

    def exit(self) -> None:
        """Exit the dialogue."""
        self.wCommand.value = ""
        self.value = ""
        self.exit_editing()

    def update_grid(self) -> None:
        if self.value:
            self.value = os.path.expanduser(self.value)
        if not os.path.exists(self.value):
            self.value = os.getcwd()
        if os.path.isdir(self.value):
            working_dir = Path(self.value)
        else:
            working_dir = Path(self.value).parent
        self.wStatus1.value = working_dir
        file_list, dir_list = [], []
        if working_dir.parent != working_dir:
            dir_list.append("..")
        try:
            for fn in working_dir.glob("*"):
                rel_path = str(fn.relative_to(working_dir))
                if not rel_path.startswith("."):
                    if fn.is_dir():
                        dir_list.append(str(fn) + os.sep)
                    elif fn.suffix in self.file_extentions:
                        file_list.append(str(fn))
        except OSError:
            npyscreen.notify_wait(
                title="Error", message="Could not read specified directory."
            )
        # DOES NOT CURRENTLY WORK - EXCEPT FOR THE WORKING DIRECTORY.  REFACTOR.
        # sort Filelist
        file_list.sort(key=str.casefold)
        dir_list.sort(key=str.casefold)
        if self.sort_by_extension:
            file_list.sort(key=self.get_extension)
        self.wMain.set_grid_values_from_flat_list(
            dir_list + file_list, reset_cursor=False
        )
        self.display()


def selectFile(starting_value: str = "", *args, **keywords):
    F = FileSelector(*args, **keywords)
    F.set_colors()
    F.wCommand.show_bold = True
    if starting_value:
        if not os.path.exists(os.path.abspath(os.path.expanduser(starting_value))):
            F.value = os.getcwd()
        else:
            F.value = starting_value
            F.wCommand.value = starting_value
    else:
        F.value = os.getcwd()
    F.update_grid()
    F.edit()
    return F.wCommand.value


class BaseForm(npyscreen.FormMultiPageWithMenus, npyscreen.FormWithMenus):
    """Base class for forms."""

    _num: int = 0
    input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = {}
    """Dictionary of input fileds: the key of the dictionary represents the name
       of the key in the in config toml input files. Values represent a tuple of
       npysceen types that display the input information on this key to the
       user and a boolean indicating whether or not this variable is mandatory.
    """
    certificates: list[str] = []
    """The type of certificate files this step needs."""

    def get_config(self, key) -> dict[str, str | bool | list[str]]:
        """Read the configuration for a step."""
        try:
            cfg = self.parentApp.config[key].copy()
            if not isinstance(cfg, dict):
                cfg = {"config": {}}
        except (KeyError, AttributeError, TypeError):
            cfg = {"config": {}}
        cfg.setdefault("config", {})
        for k, values in cfg["config"].items():
            if values is None:
                cfg["config"][k] = ""
        return cfg["config"]

    def get_host(self, key) -> str:
        """Read the host name(s) from the main windows config."""
        try:
            host = self.parentApp.config[key]["hosts"]
        except (TypeError, KeyError):
            return ""
        if isinstance(host, str):
            host = [v.strip() for v in host.split(",") if v.strip()]
        return ",".join(host)

    @property
    def num(self) -> str:
        """Calculate the number for enumerations of any input field."""
        self._num += 1
        return f"{self._num}. "

    def check_config(
        self,
        notify: bool = True,
    ) -> dict[str, str | dict[str, str | list | int | bool | None]] | None:
        """Check if the from entries are valid."""
        config = {}
        for key, (obj, mandatory) in self.input_fields.items():
            try:
                value = obj.values[obj.value]
            except AttributeError:
                value = obj.value
            if isinstance(value, str):
                if not value and self.use.value and mandatory and notify:
                    msg = f"MISSING ENTRY FOR {self.step}: {obj.name}"
                    npyscreen.notify_confirm(msg, title="ERROR")
                    return None
                elif not value and not mandatory:
                    continue
            config[key] = value
        cfg = dict(hosts=config.pop("hosts"))
        cfg["config"] = config
        for key, value in cfg["config"].items():
            if key in self.list_keys:
                cfg["config"][key] = value.split(",")
        return cfg

    def draw_form(self) -> None:
        """Overload the draw_from method from, this is done to add menus."""
        super().draw_form()
        menus = [
            " " + self.__class__.MENU_KEY + ":Main Menu ",
            "^K:Prev. Tab ",
            "^L:Next Tab ",
            "^R:Run ",
            "^S:Save ",
            "^O:Load ",
            "^E:Exit ",
        ]
        y, x = self.display_menu_advert_at()
        for n, menu_advert in enumerate(menus):
            if isinstance(menu_advert, bytes):
                menu_advert = menu_advert.decode("utf-8", "replace")
            X = sum([len(menu) for menu in menus[:n]]) + x
            self.add_line(
                y,
                X,
                menu_advert,
                self.make_attributes_list(menu_advert, curses.A_NORMAL),
                self.columns - X - 1,
            )

    def previews_form(self, *args, **keywords) -> None:
        return self.change_forms(*args, reverse=True, **keywords)

    def change_forms(self, *args, reverse=False, **keywords) -> None:
        """Cycle between the deployment config forms."""
        for step in self.parentApp._steps_lookup.keys():
            if self.name.lower().startswith(step):
                name = step
                break
            elif self.name.lower().startswith("database"):
                name = "db"
                break
        keys = self.parentApp._steps_lookup.keys()
        steps = list(self.parentApp._steps_lookup.values())
        if reverse:
            change_to = dict(zip(keys, [steps[-1]] + steps[:-1]))
        else:
            change_to = dict(zip(keys, steps[1:] + [steps[0]]))
        # raise ValueError(change_to, reverse, self.parentApp._steps_lookup)
        self.parentApp.current_form = name
        self.parentApp.change_form(change_to[name])

    def run_deployment(self, *args) -> None:
        """Switch to the deployment setup form."""
        self.parentApp.current_form = self.name
        self.parentApp.change_form("SETUP")

    def create(self) -> None:
        """Setup the form."""
        self.how_exited_handers[
            npyscreen.wgwidget.EXITED_ESCAPE
        ] = self.parentApp.exit_application
        self.add_handlers({"^O": self.parentApp.load_dialog})
        self.add_handlers({"^S": self.parentApp.save_dialog})
        self.add_handlers({"^K": self.previews_form})
        self.add_handlers({"^L": self.change_forms})
        self.add_handlers({"^R": self.run_deployment})
        self.add_handlers({"^E": self.parentApp.exit_application})
        # The menus are created here.
        self.menu = self.add_menu(name="Main Menu", shortcut="^M")
        self.submenu = self.menu.addNewSubmenu("Deployment Menu", "^D")
        self._change_form = self.parentApp.change_form
        self.menu.addItemsFromList(
            [
                ("Save Config", self.parentApp.save_dialog, "^S"),
                ("Load Config", self.parentApp.load_dialog, "^L"),
                ("Run Deployment", self.run_deployment, "^R"),
                ("Exit Application", self.parentApp.exit_application, "^E"),
            ]
        )
        self.submenu.addItemsFromList(
            [
                ("Core Deployment", self._change_form, "c", "c", ("MAIN",)),
                ("Web Deployment", self._change_form, "w", "w", ("SECOND",)),
                ("DB Deployment", self._change_form, "d", "d", ("THIRD",)),
                ("Solr Deployment", self._change_form, "s", "s", ("FOURTH",)),
                ("Run Deployment", self.run_deployment, "^R"),
            ]
        )
        self.use = self.add(
            npyscreen.CheckBox,
            max_height=2,
            value=self.step in self.parentApp._steps,
            editable=True,
            name="Check to set up the {}".format(self.step),
            scroll_exit=True,
        )
        self._add_widgets()



#!/usr/bin/python3

from subprocess import run, PIPE
import shlex
import os

if __name__ == "__main__":

    cmd = shlex.split("vault operator init")
    env = os.environ.copy()
    env["VAULT_ADDR"] = "https://127.0.0.1:8200"
    env["VAULT_SKIP_VERIFY"] = "true"
    res = run(cmd, stdout=PIPE, stderr=PIPE, env=env)
    stdout = res.stdout.decode().split("\n")
    unseal_keys = []
    print(res.stderr.decode())
    print("\n".join(stdout))
    for line in stdout:
        if "unseal key" in line.lower():
            unseal_keys.append(line.split(":")[-1].strip())
        if "initial root token" in line.lower():
            token = line.split(":")[-1].strip()
            break
    with open("/vault/keys", "w") as f:
        for nn, key in enumerate(unseal_keys):
            f.write(f"Unseal key #{nn+1}: {key}\n")
        f.write(f"Root token: {token}")

    for key in unseal_keys[:3]:
        run(shlex.split(f"vault operator unseal {key}"), env=env)
    run(shlex.split(f"vault login {token}"), env=env)
    run(shlex.split(f"vault secrets enable -path=kv kv"), env=env)
    run(shlex.split(f"vault policy write read-eval /vault/policy-file.hcl"), env=env)
    run(shlex.split(f"vault auth enable cert"), env=env)
    run(
        shlex.split(
            "vault write auth/cert/certs/app policies=web,prod,file,read-eval certificate=@/mnt/freva.crt"
        ),
        env=env,
    )



"""CLI to interact with running services."""
from __future__ import annotations
import argparse
from getpass import getuser
from pathlib import Path
import shlex
from subprocess import run, PIPE
import sys
from tempfile import TemporaryDirectory

from numpy import sign
import requests
from rich.console import Console
import yaml

from freva_deployment import __version__
from ..utils import (
    download_server_map,
    logger,
    set_log_level,
    ServiceInfo,
    get_setup_for_service,
    guess_map_server,
)


def parse_args(args: list[str]) -> argparse.Namespace:
    """Consturct command line argument parser."""
    app = argparse.ArgumentParser(
        prog="freva-service",
        description="Interact with installed freva services.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    app.add_argument(
        "command",
        type=str,
        choices=("start", "stop", "restart", "status"),
        help="The start|stop|restart|status command for the service",
    )
    app.add_argument(
        "project_name",
        type=str,
        help="Name of the project",
        default="all",
        nargs="?",
    )
    app.add_argument(
        "--server-map",
        type=str,
        default=None,
        help=(
            "Hostname of the service mapping the freva server "
            "archtiecture, Note: you can create a server map by "
            "running the deploy-freva-map command"
        ),
    )
    app.add_argument(
        "--services",
        type=str,
        nargs="+",
        default=["solr", "db", "web"],
        choices=["web", "db", "solr"],
        help="The services to be started|stopped|restarted|checked",
    )
    app.add_argument(
        "--user", "-u", type=str, help="connect as this user", default=None
    )
    app.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    app.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    argsp = app.parse_args(args)
    if "db" in argsp.services:
        argsp.services.append("vault")
    return argsp


YAML_TASK = """---
- hosts: {group}
  vars:
    ansible_python_interpreter: {python_env}
    map_server: {map_server}
  tasks:
  - name: {command}ing {service} service on {hosts}
    shell: systemctl {command} {project_name}-{service}.service
    become: yes
"""
YAML_STATUS_TASK = """---
- hosts: {group}
  vars:
    ansible_python_interpreter: {python_env}
    container: {project_name}-{service}
  tasks:
  - name: {command} {service} service on {hosts}
    shell: curl -d "{{{{ item }}}}" {map_server}/{project_name}/{service} -X PUT
    with_items:
        - mem=$(/usr/local/bin/docker-or-podman stats --no-stream {project_name}-{service} --format '{{% raw %}}{{{{.MemPerc}}}}{{% endraw %}}'|tail)
        - cpu=$(/usr/local/bin/docker-or-podman stats --no-stream {project_name}-{service} --format '{{% raw %}}{{{{.CPUPerc}}}}{{% endraw %}}'|tail)
        - status=$(/usr/local/bin/docker-or-podman ps  -a --filter name={project_name}-{service} --format '{{% raw %}}{{{{.Status}}}}{{% endraw %}}')
    become: yes

"""

YAML_INVENTORY = """
{group}:
    hosts: {hosts}
"""


def _get_playbook_for_project(
    map_server: str,
    project_name: str,
    command: str,
    services: list[str],
    config: list[ServiceInfo],
) -> tuple[list[str], list[str]]:
    """Create a ansible playbook for given services."""
    playbooks: list[str] = []
    inventory: list[str] = []
    for service in services:
        try:
            python_env, hosts = get_setup_for_service(service, config)
        except KeyError:
            continue
        group = project_name.replace("-", "_") + "_" + service
        if command == "status":
            cmd = "checking"
            tmpl = YAML_STATUS_TASK
        else:
            cmd = command
            tmpl = YAML_TASK
        playbooks += yaml.safe_load(
            tmpl.format(
                python_env=python_env,
                project_name=project_name,
                service=service,
                command=cmd,
                group=group,
                hosts=hosts,
                map_server=map_server,
            )
        )
        inventory.append(
            YAML_INVENTORY.format(
                hosts=hosts,
                project_name=project_name,
                service=service,
                group=group,
            )
        )
    return playbooks, inventory


def cli(argv: list[str] | None = None) -> None:
    """Run command line interface."""
    argv = argv or sys.argv[1:]
    argp = parse_args(argv)
    map_server = guess_map_server(argp.server_map)
    set_log_level(argp.verbose)
    user = argp.user or getuser()
    cfg = download_server_map(map_server)
    if argp.project_name.lower() == "all":
        project_names = list(cfg.keys())
    else:
        project_names = [argp.project_name]
    playbooks, inventory = [], []
    for project_name in project_names:
        try:
            config = cfg[project_name]
        except KeyError:
            logger.error("poject name %s not defined", project_name)
            continue
        try:
            p_book, i_nventory = _get_playbook_for_project(
                map_server, project_name, argp.command, argp.services, config
            )
            playbooks += p_book
            inventory += i_nventory
        except (KeyError, AttributeError) as error:
            logger.error(error.__str__())
            continue
    if not playbooks:
        return
    with TemporaryDirectory() as temp_dir:
        inventory_file = Path(temp_dir) / "inventory.yaml"
        task_file = Path(temp_dir) / "tasks.yaml"
        logger.debug("\n".join(inventory))
        with open(inventory_file, "w") as f_obj:
            f_obj.write("\n".join(inventory))
        with open(task_file, "w") as f_obj:
            yaml.dump(playbooks, f_obj)
        logger.debug(yaml.dump(playbooks))
        verbosity = sign(argp.verbose) * "-" + argp.verbose * "v"
        cmd = shlex.split(
            f"ansible-playbook {verbosity} -u {user} --ask-pass --ask-become -i "
            f"{inventory_file} {task_file}"
        )
        if verbosity:
            run(cmd, check=True)
        else:
            run(cmd, check=True, stdout=PIPE)
    if argp.command == "status":
        con = Console()
        header = ["SERVICE", "MEM", "CPU", "STATUS"]
        format_row = "{:<1} {:<20} {:^10} {:^10} {:>10}"
        con.print(format_row.format("", *header), style="bold", markup=True)
        for project in project_names:
            for service in argp.services:
                req = requests.get(f"http://{map_server}/{project}/{service}")
                status = req.json()
                status_line = [f"{project}-{service}"]
                if req.status_code != 200:
                    logger.warning("Service not found: %s", f"{project}-{service}")
                    continue
                for key in ("mem", "cpu", "status"):
                    status_line.append(status.get(key, "NaN"))
                con.print(format_row.format("", *status_line), markup=True)



#!/usr/bin/env python3
import argparse
import shutil
import sys
import subprocess
from typing import Dict, List, Tuple

SYSTEMD_TMPL = dict(
    Unit=dict(
        Description="Start/Stop freva services containers",
        After="network-online.target",
    ),
    Service=dict(
        TimeoutStartSec="35s",
        TimeoutStopSec="35s",
        ExecStart="{container_cmd} start -a {container_name}",
        ExecStop="{container_cmd} stop {container_name}",
        Restart="no",
    ),
    Install=dict(WantedBy="default.target"),
)


def parse_args() -> Tuple[str, List[str], bool]:
    """Parse the commandline arguments."""

    app = argparse.ArgumentParser(
        prog=sys.argv[0], description="Create a new systemd unit."
    )
    app.add_argument(
        "unit_name", type=str, help="Name of the systemd unit that is created."
    )
    app.add_argument(
        "--enable", action="store_true", default=False, help="Enable this unit."
    )
    app.add_argument("--requires", type=str, nargs="+", default=[])
    args = app.parse_args()
    return args.unit_name, args.requires, args.enable


def _parse_dict(tmp_dict: Dict[str, Dict[str, str]]) -> str:
    systemd_unit = ""
    for section, keys in tmp_dict.items():
        systemd_unit += "[{}]\n".format(section)
        for key, values in keys.items():
            systemd_unit += "{} = {}\n".format(key, values)
    return systemd_unit


def load_unit(unit: str, content: str, enable: bool = True) -> None:
    """Load a given systemd unit."""
    with open("/etc/systemd/system/{}.service".format(unit), "w") as f_obj:
        f_obj.write(content)
    subprocess.run(["systemctl", "daemon-reload"], check=True)
    if enable:
        subprocess.run(["systemctl", "enable", unit], check=True)
    subprocess.run(["systemctl", "restart", unit], check=True)


def create_unit(unit_name: str, requires: List[str], enable: bool) -> None:
    """Create the systemd unit."""
    container_cmd = shutil.which("podman") or shutil.which("docker")
    if container_cmd is None:
        raise ValueError("Docker or podman must be installed on the host.")
    if "docker" in container_cmd:
        SYSTEMD_TMPL["Unit"]["Requires"] = "docker.service"
        SYSTEMD_TMPL["Unit"]["After"] += " docker.service"
    for key in "ExecStop", "ExecStart":
        SYSTEMD_TMPL["Service"][key] = SYSTEMD_TMPL["Service"][key].format(
            container_cmd=container_cmd, container_name=unit_name
        )
    for service in requires:
        for key in ("After", "Requires"):
            try:
                SYSTEMD_TMPL["Unit"][key] += " {}.service".format(service)
            except KeyError:
                SYSTEMD_TMPL["Unit"][key] = " {}.service".format(service)
    load_unit(unit_name, _parse_dict(SYSTEMD_TMPL), enable)


if __name__ == "__main__":
    create_unit(*parse_args())



"""Collection of utils for deployment."""
from __future__ import annotations
import hashlib
import logging
import json
from pathlib import Path
import re
from subprocess import PIPE
import sys
import shutil
from typing import cast, NamedTuple

import appdirs
import requests
from rich.console import Console
from rich.prompt import Prompt
import toml

logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s", level=logging.INFO)
logger = logging.getLogger("freva-deployment")

RichConsole = Console(markup=True, force_terminal=True)

password_prompt = (
    "[green]Choose[/] a [b]master password[/], this password will be used to:\n"
    "- create the [magenta]mysql root[/] password\n"
    "- set the [magenta]django admin[/] web password\n"
    "[b]Note:[/] Ideally this password can be shared amongst other [i]admins[/].\n"
    "[b][green]choose[/] master password[/]"
)

ServiceInfo = NamedTuple(
    "ServiceInfo", [("name", str), ("python_path", str), ("hosts", str)]
)


class AssetDir:
    def __init__(self):

        self._user_asset_dir = Path(appdirs.user_data_dir()) / "freva" / "deployment"
        self._user_config_dir = Path(appdirs.user_config_dir()) / "freva" / "deployment"
        self._central_asset_dir = Path(sys.exec_prefix) / "freva" / "deployment"

    @property
    def asset_dir(self):
        if self._user_asset_dir.exists():
            return self._user_asset_dir
        return self._central_asset_dir

    @property
    def config_dir(self):
        inventory_file = self._user_config_dir / "inventory.toml"
        if inventory_file.exists():
            return self._user_config_dir
        self._user_config_dir.mkdir(exist_ok=True, parents=True)
        shutil.copy(self.asset_dir / "config" / "inventory.toml", inventory_file)
        return self._user_config_dir


AD = AssetDir()
asset_dir = AD.asset_dir
config_dir = AD.config_dir


def guess_map_server(
    inp_server: str | None, default_port: int = 6111, mandatory: bool = True
) -> str:
    """Try to get the server name mapping the freva infrastructure.

    Parameters
    ----------
    inp_server: str | None
        Input server name, if None given, the code tries to read the
        the server name from a previous deployment setup.
    default_port: int, default: 6111
        The port to connect to
    mandatory: bool, default: True
        If mandatory is set and no server could be found an error is risen.

    Returns
    -------
    str: hostname of the server that runs the server map service

    Raises
    ------
    ValueError: If no server could be found a ValueError is raised.
    """
    inp_server = inp_server or ""
    if inp_server:
        host_name, _, port = inp_server.partition(":")
        port = port or str(default_port)
        return f"{host_name}:{port}"
    cache_dir = Path(appdirs.user_cache_dir()) / "freva-deployment"
    try:
        with (cache_dir / "freva_deployment.json").open() as f_obj:
            inp_server = cast(str, json.load(f_obj)["server_map"])
            host_name, _, port = inp_server.partition(":")
            port = port or str(default_port)
            return f"{host_name}:{port}"
    except (FileNotFoundError, IOError, ValueError, KeyError, json.JSONDecodeError):
        pass
    if mandatory:
        raise ValueError(
            "You have to set the hostname to the services mapping "
            "the freva infrastructure using the --server-map flag."
        )
    return ""


def set_log_level(verbosity: int) -> None:
    """Set the log level of the logger."""
    logger.setLevel(max(logging.INFO - 10 * verbosity, logging.DEBUG))


def get_setup_for_service(service: str, setups: list[ServiceInfo]) -> tuple[str, str]:
    """Get the setup of a service configuration."""
    for setup in setups:
        if setup.name == service:
            return setup.python_path, setup.hosts
    raise KeyError("Service not found")


def read_db_credentials(
    cert_file: Path, db_host: str, port: int = 5002
) -> dict[str, str]:
    """Read database config."""
    with cert_file.open() as f_obj:
        key = "".join([k.strip() for k in f_obj.readlines() if not k.startswith("-")])
        sha = hashlib.sha512(key.encode()).hexdigest()
    url = f"http://{db_host}:{port}/vault/data/{sha}"
    return requests.get(url).json()


def download_server_map(
    server_map,
) -> dict[str, list[ServiceInfo]]:
    """Download server information from the service that stores the server arch.

    Parameters
    ----------
    server_map: str,
        The hostname holding the server archtiecture information.
    Returns
    -------
    dict[str, list[NamedTuple("service", [("name", str), ("python_path", str),
                                     ("hosts", str)])]]
    """
    info: dict[str, list[ServiceInfo]] = {}
    host, _, port = server_map.partition(":")
    port = port or "6111"
    req = requests.get(f"http://{host}:{port}")
    if req.status_code != 200:
        logger.error(req.json())
        return {}
    for proj, conf in req.json().items():
        info[cast(str, proj)] = [
            ServiceInfo(name=s, python_path=c[0], hosts=c[-1])
            for (s, c) in conf.items()
        ]
    return info


def upload_server_map(
    server_map: str,
    project_name: str,
    deployment_setup: dict[str, dict[str, str]],
) -> None:
    """Upload server information to service that stores server archtiecture.

    Parameters
    ----------
    server_map: str
        The hostname holding the server archtiecture information.
    project_name: str
        Name of the freva project
    deployment_setup: dict[str, str]
        Server names for each deployed service
    """
    _upload_data: dict[str, dict[str, str]] = {}
    for service, config in deployment_setup.items():
        _upload_data[service] = config
    host, _, port = server_map.partition(":")
    port = port or "6111"
    req_data = dict(config=toml.dumps(_upload_data))
    logger.debug("Uploading %s", req_data)
    req = requests.put(f"http://{host}:{port}/{project_name}", data=req_data)
    if req.status_code == 201:
        logger.info("Server information updated at %s", host)
    else:
        logger.error("Could not update server information %s", req.json())


def get_email_credentials() -> tuple[str, str]:
    """Read login credentials for email server.

    Returns
    =======
    tuple: username and password
    """

    msg = (
        "\nThe web will need login credentials to connect to the [b green]mail server [/]"
        "that has been set up.\nYou should now enter your [it]login credentials[/].\n"
        "[b]Note:[/]These credentials will be securely stored in an encrypted vault\n"
    )
    RichConsole.print(msg)
    username = Prompt.ask("[green b]Username[/] for mail server")
    password = Prompt.ask("[green b]Password[/] for mail server", password=True)
    return username, password


def get_passwd(min_characters: int = 8) -> str:
    """Create a secure pasword.

    Parameters
    ==========
    min_characters:
        The minimum lenght of the password (default 8)

    Returns
    =======
    str: The password
    """
    logger.info("Creating Password")
    msg = ""
    RichConsole
    while True:
        try:
            return _create_passwd(min_characters, msg)
        except ValueError as e:
            RichConsole.print(f"[red]{e.__str__()}[/]")
            msg = "[b red]re-enter[/] master password"


def _create_passwd(min_characters: int, msg: str = "") -> str:
    """Create passwords."""
    master_pass = Prompt.ask(msg or password_prompt, password=True)
    is_ok: bool = len(master_pass) > min_characters
    for check in ("[a-z]", "[A-Z]", "[0-9]"):
        if not re.search(check, master_pass):
            is_ok = False
            break
    is_safe: bool = len([True for c in "[_@$#$%^&*-!]" if c in master_pass]) > 0
    if is_ok is False or is_safe is False:
        raise ValueError(
            (
                "Password must confirm the following constraints:\n"
                f"- {min_characters} alphanumeric characters long,\n"
                "- have both, lower and upper case characters,\n"
                "- have at least one special special character."
            )
        )
    master_pass_2 = Prompt.ask("[bold green]re-enter[/] master password", password=True)
    if master_pass != master_pass_2:
        raise ValueError("Passwords do not match")
    return master_pass



__version__ = "2208.0.1"
AVAILABLE_PYTHON_VERSIONS = ["3.7", "3.8", "3.9", "3.10"]
AVAILABLE_CONDA_ARCHS = [
    "Linux-x86_64",
    "Linux-aarch64",
    "Linux-ppc64le",
    "Linux-s390x",
    "MacOSX-x86_64",
]



#!/usr/bin/env python3

import os
import random
import shlex
from subprocess import run
import string

import requests

SQL_ENTRY = (
    "mysql -h {host} -u root -p'{root_pw}' -e "
    "\"FLUSH PRIVILEGES; ALTER USER {user}@'%' IDENTIFIED BY "
    "'{passwd}'; FLUSH PRIVILEGES\""
)


def gen_passwd(
    num_chars: int = 20, num_digits: int = 4, num_punctuations: int = 4
) -> str:
    """Generate a new mysql db user password."""

    num_chars -= num_digits + num_punctuations
    punctuations = "!@$^&*()_+-;:|,.%"
    possible_characters = [
        "".join(random.sample(string.ascii_letters, num_chars)),
        "".join(random.sample(string.digits, num_digits)),
        "".join(random.sample(punctuations, num_punctuations)),
    ]
    characters = "".join(possible_characters)
    return "".join(random.sample(characters, len(characters)))


def set_passwd_in_sql_server(passwd: str) -> None:
    """Set the new password of the sql user."""

    root_pw = os.environ["MYSQL_ROOT_PASSWORD"]
    url = f'http://{os.environ["HOST"]}:5002/vault/{passwd}/{root_pw}'
    cmd = SQL_ENTRY.format(
        user=os.environ["DB_USER"],
        passwd=passwd,
        host="localhost",
        root_pw=root_pw,
    )
    _ = run(shlex.split(cmd), check=True)
    requests.post(url).json()


if __name__ == "__main__":
    set_passwd_in_sql_server(gen_passwd())



from __future__ import annotations
from getpass import getuser
import npyscreen
from pathlib import Path
from typing import cast, List, Dict

from .base import BaseForm, logger
from freva_deployment import AVAILABLE_PYTHON_VERSIONS, AVAILABLE_CONDA_ARCHS


def get_index(values: list[str], target: str, default: int = 0) -> int:
    """Get the target index of item in list.

    Parameters:
    ===========
    values:
        the list of values that is searched
    target:
        the item the list of values that is searched for
    default:
        if nothing is found return the default value

    Returns:
    ========
    int: Index of the the target item in the list
    """

    for n, value in enumerate(values):
        if value == target:
            return n
    return default


class CoreScreen(BaseForm):
    """Form for the core deployment configuration."""

    step: str = "core"
    """Name of this step."""
    certificates: list[str] = ["public"]
    """The type of certificate files this step needs."""

    @property
    def scheduler_systems(self):
        """Define available scheduler systems."""
        return ["local", "slurm", "pbs", "lfs", "moab", "oar", "sge"]

    def scheduler_index(self, scheduler_system: str = ""):
        """Get the index of the saved scheduler_system"""
        scheduler_system = scheduler_system or "local"
        for nn, choice in enumerate(self.scheduler_systems):
            if choice == scheduler_system:
                return nn
        return 0

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        arch = cast(str, cfg.get("arch", AVAILABLE_CONDA_ARCHS[0]))
        arch_idx = get_index(AVAILABLE_CONDA_ARCHS, arch, 0)
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            hosts=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Server Name(s) where core is deployed:",
                    value=self.get_host("core"),
                ),
                True,
            ),
            install_dir=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Anaconda installation dir. for core:",
                    value=cfg.get("install_dir", ""),
                ),
                True,
            ),
            install=(
                self.add_widget_intelligent(
                    npyscreen.RoundCheckBox,
                    max_height=2,
                    value=cfg.get("install", True),
                    editable=True,
                    name=(f"{self.num}Install a new Freva anaconda environment?"),
                    scroll_exit=True,
                ),
                True,
            ),
            root_dir=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=(
                        f"{self.num}Directory where project configuration is stored "
                        "(defaults to `anaconda installation dir.` in #3):"
                    ),
                    value=cfg.get("root_dir", ""),
                ),
                False,
            ),
            base_dir_location=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}User data directory:",
                    value=cfg.get("base_dir_location", ""),
                ),
                False,
            ),
            preview_path=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Plugin output dir for the web UI (preview path):",
                    value=cfg.get("preview_path", ""),
                ),
                False,
            ),
            scheduler_system=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=f"{self.num}Workload manger",
                    value=self.scheduler_index(cast(str, cfg.get("scheduler_system"))),
                    values=self.scheduler_systems,
                ),
                True,
            ),
            scheduler_output_dir=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Ouput dir. of the scheduler system, "
                    "${base_dir_location}/share",
                    value=cfg.get("scheduler_output_dir", ""),
                ),
                False,
            ),
            admins=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Set the admin user(s) - comma separated:",
                    value=cfg.get("admins", getuser()),
                ),
                False,
            ),
            admin_group=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(
                        f"{self.num}Set the Freva admin group - "
                        "leave blank if not needed:"
                    ),
                    value=cfg.get("admin_group", ""),
                ),
                False,
            ),
            ansible_become_user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(
                        f"{self.num}Optional username to change to on the "
                        "remote machine, leave blank for none:"
                    ),
                    value=cfg.get("ansible_become_user", ""),
                ),
                False,
            ),
            conda_exec_path=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(
                        f"{self.num}Path to any existing conda installation - "
                        "leave blank to "
                        "install a temporary conda distribution:"
                    ),
                    value=cfg.get("conda_exec_path", ""),
                ),
                False,
            ),
            arch=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=(
                        f"{self.num}Set the target architecture of the system where "
                        "the backend will be installed:"
                    ),
                    value=arch_idx,
                    values=AVAILABLE_CONDA_ARCHS,
                ),
                True,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Python path on remote machine:",
                    value=cfg.get("ansible_python_interpreter", "/usr/bin/python3"),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Username for remote machine:",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
            git_path=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Path to the git executable (leave blank for default):",
                    value=cfg.get("git_path", ""),
                ),
                False,
            ),
        )


class WebScreen(BaseForm):
    """Form for the web deployment configuration."""

    step: str = "web"
    certificates: list[str] = ["public", "private", "chain"]
    """The type of certificate files this step needs."""

    def get_index(self, choices: list[str], key: str):
        """Get the key value pair for a combo box"""
        for nn, choice in enumerate(choices):
            if choice == key:
                return nn
        return 0

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys = "imprint", "scheduler_host"
        cfg = self.get_config(self.step)
        for key in self.list_keys:
            if key in cfg and isinstance(cfg[key], str):
                value = cast(str, cfg[key])
                cfg[key] = [v.strip() for v in value.split(",") if v.strip()]
                logger.warning(key, cfg[key])
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            hosts=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Server Name(s) the web service is deployed on:",
                    value=self.get_host("web"),
                ),
                True,
            ),
            project_website=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Url of the Freva home page:",
                    value=cfg.get("project_website", ""),
                ),
                True,
            ),
            main_color=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Html color of the main color theme:",
                    value=cfg.get("main_color", "Tomato"),
                ),
                True,
            ),
            border_color=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Html color for the borders:",
                    value=cfg.get("border_color", "#6c2e1f"),
                ),
                True,
            ),
            hover_color=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Html color for hover modes:",
                    value=cfg.get("hover_color", "#d0513a"),
                ),
                True,
            ),
            institution_logo=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Path to the institution logo.",
                    value=cfg.get(
                        "institution_logo", "/path/to/logo/on/target/machine"
                    ),
                ),
                True,
            ),
            about_us_text=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}About us text - short blurb about Freva:",
                    value=cfg.get("about_us_text", "Testing"),
                ),
                True,
            ),
            contacts=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Contact email address:",
                    value=str(cfg.get("contacts", "data@dkrz.de")),
                ),
                True,
            ),
            email_host=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(
                        f"{self.num}Smtp email that will be used to send "
                        "emails to the contacts via the web ui"
                    ),
                    value=cfg.get("email_host", "mailhost.dkrz.de"),
                ),
                True,
            ),
            imprint=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Institution address - comma separated:",
                    value=",".join(
                        cast(
                            List[str],
                            cfg.get(
                                "imprint",
                                [
                                    "freva",
                                    "German Climate Computing Centre (DKRZ)",
                                    "Bundesstr. 45a",
                                    "20146 Hamburg",
                                    "Germany",
                                ],
                            ),
                        ),
                    ),
                ),
                True,
            ),
            homepage_text=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}More in detail project description:",
                    value=cfg.get(
                        "homepage_text",
                        (
                            "Lorem ipsum dolor sit amet, consectetur "
                            "adipiscing elit, sed do eiusmod tempor "
                            "incididunt ut labore et dolore magna aliqua. "
                            "Ut enim ad minim veniam, quis nostrud "
                            "exercitation ullamco laboris nisi ut aliquip "
                            "ex ea commodo consequat. Duis aute irure dolor "
                            "in reprehenderit in voluptate velit esse cillum "
                            "dolore eu fugiat nulla pariatur. Excepteur sint "
                            "occaecat cupidatat non proident, sunt in culpa "
                            "qui officia deserunt mollit anim id est laborum."
                        ),
                    ),
                ),
                True,
            ),
            homepage_heading=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}A brief describtion of the project:",
                    value=cfg.get("homepage_heading", "Lorem ipsum dolor sit amet"),
                ),
                True,
            ),
            scheduler_host=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Scheduler hostname(s) - comma separated:",
                    value=",".join(
                        cast(List[str], cfg.get("scheduler_host", ["levante.dkrz.de"]))
                    ),
                ),
                True,
            ),
            auth_ldap_server_uri=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(
                        f"{self.num}Ldap server name(s) used for authentication"
                        " - comma separated:"
                    ),
                    value=cfg.get(
                        "auth_ldap_server_uri",
                        "ldap://idm-dmz.dkrz.de",
                    ),
                ),
                True,
            ),
            auth_ldap_start_tls=(
                self.add_widget_intelligent(
                    npyscreen.RoundCheckBox,
                    max_height=2,
                    value=cfg.get("auth_ldap_start_tls", False),
                    editable=True,
                    name=(
                        f"{self.num}Enable TLS encryption when communicating with the"
                        "ldap server. Needs to be configured:"
                    ),
                    scroll_exit=True,
                ),
                True,
            ),
            allowed_group=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Unix group allowed to log on to the web:",
                    value=cfg.get("allowed_group", "my_freva"),
                ),
                True,
            ),
            ldap_user_base=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Ldap search keys for user base:",
                    value=cfg.get(
                        "ldap_user_base", "cn=users,cn=accounts,dc=dkrz,dc=de"
                    ),
                ),
                True,
            ),
            ldap_group_base=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Ldap search keys for group base:",
                    value=cfg.get(
                        "ldap_group_base",
                        "cn=groups,cn=accounts,dc=dkrz,dc=de",
                    ),
                ),
                True,
            ),
            ldap_user_dn=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Distinguished name (dn) for the ldap user:",
                    value=cfg.get(
                        "ldap_user_dn",
                        "uid=dkrzagent,cn=sysaccounts,cn=etc,dc=dkrz,dc=de",
                    ),
                ),
                True,
            ),
            ldap_user_pw=(
                self.add_widget_intelligent(
                    npyscreen.TitlePassword,
                    name=f"{self.num}Password for ldap user:",
                    value=cfg.get("ldap_user_pw", "dkrzprox"),
                ),
                True,
            ),
            ldap_first_name_field=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(f"{self.num}Ldap search search key for first name"),
                    value=cfg.get(
                        "ldap_first_name_field",
                        "givenname",
                    ),
                ),
                False,
            ),
            ldap_last_name_field=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(f"{self.num}Ldap search search key for last name"),
                    value=cfg.get(
                        "ldap_last_name_field",
                        "sn",
                    ),
                ),
                False,
            ),
            ldap_email_field=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(f"{self.num}Ldap search search key for email addr"),
                    value=cfg.get(
                        "ldap_email_field",
                        "mail",
                    ),
                ),
                False,
            ),
            ldap_group_class=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(f"{self.num}Ldap object class"),
                    value=cfg.get(
                        "ldap_group_class",
                        "groupOfNames",
                    ),
                ),
                True,
            ),
            ldap_group_type=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=(f"{self.num}Ldap group type"),
                    value=self.get_index(
                        ["posix", "nested"],
                        cast(str, cfg.get("ldap_group_type", "nested")),
                    ),
                    values=["posix", "nested"],
                ),
                True,
            ),
            ldap_model=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=(f"{self.num}Ldap tools class to be used for authentication."),
                    value=cfg.get("ldap_model", "MiklipUserInformation"),
                ),
                True,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Pythonpath on remote machine:",
                    value=cfg.get("ansible_python_interpreter", "/usr/bin/python3"),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Username for remote machine:",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class DBScreen(BaseForm):
    """Form for the core deployment configuration."""

    step: str = "db"

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        db_ports: list[int] = list(range(3300, 3320))
        port_idx = get_index(
            cast(List[str], list(map(str, db_ports))), str(cfg.get("port", 3306)), 6
        )
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            hosts=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Server Name(s) where the database service is deployed:",
                    value=self.get_host("db"),
                ),
                True,
            ),
            wipe=(
                self.add_widget_intelligent(
                    npyscreen.RoundCheckBox,
                    max_height=2,
                    value=cfg.get("wipe", False),
                    editable=True,
                    name=(f"{self.num}Delete existing data?"),
                    scroll_exit=True,
                ),
                True,
            ),
            user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Database user:",
                    value=cfg.get("user", "evaluation_system"),
                ),
                True,
            ),
            db=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Database name:",
                    value=cfg.get("db", "evaluation_system"),
                ),
                True,
            ),
            port=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=f"{self.num}Database Port:",
                    value=port_idx,
                    values=db_ports,
                ),
                True,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Pythonpath on remote machine:",
                    value=cfg.get("ansible_python_interpreter", "/usr/bin/python3"),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Username for remote machine:",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class SolrScreen(BaseForm):
    """Form for the solr deployment configuration."""

    step: str = "solr"

    def _add_widgets(self) -> None:
        """Add widgets to the screen."""
        self.list_keys: list[str] = []
        cfg = self.get_config(self.step)
        solr_ports: list[int] = list(range(8980, 9000))
        port_idx = get_index(
            [str(p) for p in solr_ports], str(cfg.get("port", 8983)), 3
        )
        self.input_fields: dict[str, tuple[npyscreen.TitleText, bool]] = dict(
            hosts=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Server Name(s) where the solr service is deployed:",
                    value=self.get_host("solr"),
                ),
                True,
            ),
            wipe=(
                self.add_widget_intelligent(
                    npyscreen.RoundCheckBox,
                    max_height=2,
                    value=cfg.get("wipe", False),
                    editable=True,
                    name=(f"{self.num}Delete existing data?"),
                    scroll_exit=True,
                ),
                True,
            ),
            mem=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=f"{self.num}Virtual memory (in GB) for the solr server:",
                    value=3,
                    values=[f"{i}g" for i in range(1, 10)],
                ),
                True,
            ),
            port=(
                self.add_widget_intelligent(
                    npyscreen.TitleCombo,
                    name=f"{self.num}Solr port:",
                    value=port_idx,
                    values=solr_ports,
                ),
                True,
            ),
            ansible_python_interpreter=(
                self.add_widget_intelligent(
                    npyscreen.TitleFilename,
                    name=f"{self.num}Pythonpath on remote machine:",
                    value=cfg.get("ansible_python_interpreter", "/usr/bin/python3"),
                ),
                False,
            ),
            ansible_user=(
                self.add_widget_intelligent(
                    npyscreen.TitleText,
                    name=f"{self.num}Username for remote machine:",
                    value=cfg.get("ansible_user", getuser()),
                ),
                False,
            ),
        )


class RunForm(npyscreen.FormMultiPageAction):
    """Definition of the form that applies the actual deployment."""

    _num: int = 0

    @property
    def num(self) -> str:
        """Calculate the number for enumerations of any input field."""
        self._num += 1
        return f"{self._num}. "

    def on_ok(self) -> None:
        """Define what happens once the `ok` for applying the deployment is hit."""

        self.parentApp.thread_stop.set()
        if not self.project_name.value:
            npyscreen.notify_confirm("You have to set a project name", title="ERROR")
            return
        if not self.server_map.value:
            value = npyscreen.notify_yes_no(
                "If you don't set a map server value you wont be able "
                "to start|stop the services. Continue anyway?",
                title="WARNING",
            )
            if not value:
                return
        missing_form: None | str = self.parentApp.check_missing_config()
        if missing_form:
            self.parentApp.change_form(missing_form)
            return
        public_keyfile = self.public_keyfile.value or self.chain_keyfile.value
        cert_files = dict(
            public=public_keyfile or "",
            private=self.private_keyfile.value or "",
        )
        for key_type, keyfile in cert_files.items():
            for step, deploy_form in self.parentApp._forms.items():
                if not keyfile or not Path(keyfile).is_file():
                    if (
                        key_type in deploy_form.certificates
                        and step in self.parentApp.steps
                    ):
                        if keyfile:
                            msg = f"{key_type} certificate file `{keyfile}` must exist."
                        else:
                            msg = f"You must give a {key_type} certificate file."
                        npyscreen.notify_confirm(msg, title="ERROR")
                        return

        cert_files["chain"] = self.chain_keyfile.value or ""
        if not cert_files["chain"]:
            value = npyscreen.notify_yes_no(
                "It is advised to create a chained certificate file. "
                "This enhances the web ui security. Continue anyway?",
                title="WARNING",
            )
            if not value:
                return
        save_file = self.parentApp.save_config_to_file(write_toml_file=True)
        if isinstance(save_file, Path):
            save_file = str(save_file)
        self.parentApp.setup = {
            "server_map": self.server_map.value,
            "steps": list(set(self.parentApp.steps)),
            "ask_pass": bool(self.use_ssh_pw.value),
            "config_file": save_file or None,
        }
        self.parentApp.exit_application(msg="Do you want to continue?")

    def on_cancel(self) -> None:
        """Define what happens after the the cancel button is hit."""
        name = self.parentApp.current_form.lower()
        self.parentApp.setup = {}
        for step, form_name in self.parentApp._steps_lookup.items():
            if name.startswith(step):
                # Tell the MyTestApp object to change forms.
                self.parentApp.change_form(form_name)
                return
        self.parentApp.change_form("MAIN")

    def create(self) -> None:
        """Custom definitions executed when the from gets created."""
        self.how_exited_handers[
            npyscreen.wgwidget.EXITED_ESCAPE
        ] = self.parentApp.exit_application
        self._add_widgets()

    def _add_widgets(self) -> None:
        """Add the widgets to the form."""

        project_name = self.parentApp.config.get(
            "project_name", self.parentApp._read_cache("project_name", "")
        )
        ssh_pw = self.parentApp._read_cache("ssh_pw", True)
        self.project_name = self.add_widget_intelligent(
            npyscreen.TitleText,
            name=f"{self.num}Set the name of the project",
            value=project_name,
        )
        self.inventory_file = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Save config as",
            value=str(self.parentApp.save_file or ""),
        )
        self.server_map = self.add_widget_intelligent(
            npyscreen.TitleText,
            name=(f"{self.num}Hostname of the service mapping the freva server arch."),
            value=self.parentApp._read_cache("server_map", ""),
        )
        self.public_keyfile = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Select a public certificate file; needed for steps web, core, db",
            value=self.parentApp.read_cert_file("public_keyfile"),
        )
        self.private_keyfile = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Select a private certificate file; needed for steps web",
            value=self.parentApp.read_cert_file("private_keyfile"),
        )
        self.chain_keyfile = self.add_widget_intelligent(
            npyscreen.TitleFilename,
            name=f"{self.num}Select a chain certificate file; needed for steps web",
            value=self.parentApp.read_cert_file("chain_keyfile"),
        )
        self.use_ssh_pw = self.add_widget_intelligent(
            npyscreen.RoundCheckBox,
            max_height=2,
            value=ssh_pw,
            editable=True,
            name=f"{self.num}Use password for ssh connection",
            scroll_exit=True,
        )



#!/usr/bin/python3

from flask import Flask, request, jsonify
from flask_restful import Resource, Api
from json import dumps
from pathlib import Path
import sys
from subprocess import run, PIPE, Popen
import shlex
import time
import os
import hashlib
from functools import cached_property
import requests

app = Flask(__name__)
api = Api(app)
KEY_FILE = Path("/vault/keys")
CERT_DIR = Path("/data")


def read_key():
    unseal_keys = []
    token = ""
    with KEY_FILE.open("r") as f:
        for line in f.readlines():
            if "unseal key" in line.lower():
                unseal_keys.append(line.split(":")[-1].strip())
            if "initial root token" in line.lower():
                token = line.split(":")[-1].strip()
    return unseal_keys, token


def unseal():

    env = os.environ.copy()
    env["VAULT_ADDR"] = "http://127.0.0.1:8200"
    env["VAULT_SKIP_VERIFY"] = "true"
    unseal_keys, _ = read_key()
    for key in unseal_keys[:3]:
        run(shlex.split(f"vault operator unseal {key}"), env=env)


def deploy_vault():

    cmd = shlex.split("vault operator init")
    env = os.environ.copy()
    env["VAULT_ADDR"] = "http://127.0.0.1:8200"
    env["VAULT_SKIP_VERIFY"] = "true"
    res = run(cmd, stdout=PIPE, stderr=PIPE, env=env)
    with KEY_FILE.open("w") as f:
        f.write(res.stdout.decode())
    unseal_keys, token = read_key()
    unseal()
    if token:
        run(shlex.split(f"vault login {token}"), env=env)
        run(shlex.split(f"vault secrets enable -version=2 -path=kv kv"), env=env)
        run(
            shlex.split(f"vault policy write read-eval /vault/policy-file.hcl"), env=env
        )


class Vault(Resource):
    @staticmethod
    def _read_key(suffix):
        with (CERT_DIR / f"freva.{suffix}").open() as f:
            key = [k.strip() for k in f.readlines() if not k.startswith("-")]
        return ("".join(key)).encode()

    @cached_property
    def public_key(self):
        return hashlib.sha512(self._read_key("crt")).hexdigest()

    def post(self, entry, public_key):
        """Post method, for setting a new db_password.

        Parameters:
        -----------
        entry: str
            the new database password
        public_key: str
            mysql root password to

        Returns:
        --------
            dict: status information
        """
        out, status = {}, 401
        if public_key != os.environ["ROOT_PW"]:
            return out, status
        _, token = read_key()
        # Get the information from the vault
        url = f"http://127.0.0.1:8200/v1/kv/data/read-eval"
        headers = {"X-Vault-Token": token}
        out = requests.get(url, headers=headers).json()["data"]
        out["data"]["db.passwd"] = entry
        r = requests.post(url, json=out, headers=headers)
        return {}, 201

    def get(self, entry, public_key):
        """Get method, for getting vault information.

        Parameters:
        -----------
        entry: str
            the kind of inormation that is returned, can be
            key#N : to get the unseal key #N
            token: to get the login token
            data: to get the key-value data that is stored in the vault
        public_key: hashlib.sha512
            hexdigest representation of the sha512 public key that is stored
            on the docker container.

        Returns:
        --------
            dict: Vault information
        """
        out = {}
        status = 401
        _, token = read_key()
        if public_key != self.public_key:
            return out, 401
        if entry == "token":
            return {"X-Vault-Token": token}, 200
        if entry == "data":
            # Get the information from the vault
            req = requests.get(
                "http://127.0.0.1:8200/v1/kv/data/read-eval",
                headers={"X-Vault-Token": token},
            )
        else:
            req = requests.get(
                f"http://127.0.0.1:8200/v1/kv/data/{entry}",
                headers={"X-Vault-Token": token},
            )
            out = req.json()["data"]["data"]
            status = req.status_code
        try:
            out = req.json()["data"]["data"]
            status = req.status_code
        except KeyError:
            out = req.json()
        return out, status


api.add_resource(Vault, "/vault/<entry>/<public_key>")  # Route_3

if __name__ == "__main__":
    try:
        cmd = sys.argv[1:]
    except IndexError:
        cmd = ["vault", "server", "-config", "/vault/vault-server-tls.hcl"]
    if len(cmd) == 0:
        cmd = ["vault", "server", "-config", "/vault/vault-server-tls.hcl"]
    p = Popen(cmd, env=os.environ.copy())
    time.sleep(1)
    if not KEY_FILE.exists():
        deploy_vault()
    else:
        unseal()
    app.run(host="0.0.0.0", port="5002")
    p.wait()



#!/usr/bin/env python3
"""Setup script for packaging freva deployment."""

from pathlib import Path
import re
import urllib.request
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
import sys
from typing import List


THIS_DIR = Path(__file__).parent
CONFIG_DIR = Path("freva") / "deployment"
ASSET_DIR = THIS_DIR / "assets"


INSTALL_REQUIRES = [
    "appdirs",
    "npyscreen",
    "numpy",
    "PyMySQL",
    "pyncclient",
    "pyyml",
    "rich",
    "toml",
    "tomlkit",
    "requests",
]


def find_version(*parts):
    vers_file = read(*parts)
    match = re.search(r'^__version__ = "(\d+.\d+.\d+)"', vers_file, re.M)
    if match is not None:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


def download_assets() -> None:
    """Download additional assets."""
    assets = [
        (
            ASSET_DIR / "config" / "evaluation_system.conf.tmpl",
            "https://gitlab.dkrz.de/freva/evaluation_system/-/raw/freva-dev/assets/evaluation_system.conf",
        )
    ]
    for path, url in assets:
        urllib.request.urlretrieve(url, filename=str(path))


def prepare_config(develop_cmd: bool = False) -> None:
    """Create the freva config dir in loacal user directory."""
    import appdirs

    download_assets()
    user_config_dir = Path(appdirs.user_config_dir()) / CONFIG_DIR
    user_data_dir = Path(appdirs.user_data_dir()) / CONFIG_DIR
    for cfg_path in (user_config_dir, user_data_dir):
        cfg_path.mkdir(exist_ok=True, parents=True)
    inventory_file = user_config_dir / "inventory.toml"
    inventory_asset = ASSET_DIR / "config" / "inventory.toml"
    with inventory_asset.open() as f:
        if inventory_file.is_symlink():
            inventory_file.unlink()
        if not inventory_file.exists() and develop_cmd is False:
            with inventory_file.open("w") as w:
                w.write(f.read())
        elif develop_cmd is True:
            inventory_file.unlink(missing_ok=True)
            inventory_file.symlink_to(inventory_asset)
    for datafile in ASSET_DIR.rglob("*"):
        new_path = user_data_dir / datafile.relative_to(ASSET_DIR)
        if datafile.is_dir():
            continue
        new_path.parent.mkdir(exist_ok=True, parents=True)
        with datafile.open() as f:
            new_path.unlink(missing_ok=True)
            if develop_cmd is False:
                with new_path.open("w") as w:
                    w.write(f.read())
            else:
                new_path.symlink_to(datafile)


class InstallCommand(install):
    """Customized setuptools install command."""

    def run(self):
        install.run(self)
        prepare_config(develop_cmd=False)


class DevelopCommand(develop):
    """Customized setuptools install command."""

    def run(self):
        develop.run(self)
        prepare_config(develop_cmd=True)


def read(*parts: str) -> str:
    """Read the content of a file."""
    with THIS_DIR.joinpath(*parts).open() as f:
        return f.read()


def get_packages() -> List[str]:
    """Get the packages needed to install."""
    plf = sys.platform
    if plf.startswith("win") or plf.startswith("cygwin") or plf.startswith("ms"):
        return INSTALL_REQUIRES
    INSTALL_REQUIRES.append("ansible")
    return INSTALL_REQUIRES


def get_data_files() -> List[str]:
    dirs = [d for d in ASSET_DIR.rglob("*") if d.is_dir()]
    files = []
    for d in dirs:
        target_dir = Path("share") / "freva" / "deployment" / d.relative_to(ASSET_DIR)
        add_files = [str(f.relative_to(THIS_DIR)) for f in d.rglob("*") if f.is_file()]
        if add_files:
            files.append((str(target_dir), add_files))
    return files


setup(
    name="freva_deployment",
    version=find_version("src", "freva_deployment", "__init__.py"),
    author="Martin Bergemann",
    author_email="martin.bergemann@dkrz.de",
    maintainer="Martin Bergemann",
    url="https://gitlab.dkrz.de/freva/deployment.git",
    description="Deploy freva and its services on different machines.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    include_package_data=True,
    data_files=get_data_files(),
    license="GPLv3",
    packages=find_packages("src"),
    package_dir={"": "src"},
    cmdclass={
        "develop": DevelopCommand,
        "install": InstallCommand,
    },
    entry_points={
        "console_scripts": [
            "deploy-freva-cmd = freva_deployment.cli:deploy",
            "deploy-freva-map = freva_deployment.cli:server_map",
            "freva-service = freva_deployment.cli:service",
            "freva-migrate = freva_deployment.cli:migrate",
            "deploy-freva = freva_deployment.ui.deployment_tui:tui",
        ]
    },
    setup_requires=["appdirs"],
    install_requires=get_packages(),
    extras_require={
        "docs": [
            "sphinx",
            "nbsphinx",
            "recommonmark",
            "sphinx_rtd_theme",
            "ipython",  # For nbsphinx syntax highlighting
            "sphinxcontrib_github_alt",
        ],
        "test": ["mypy", "black", "types-toml", "types-PyMySQL"],
    },
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ],
)



"""CLI to assist with migrating the system."""
from __future__ import annotations
import argparse
import json
from pathlib import Path
import os
import shlex
import shutil
from subprocess import run, PIPE, CalledProcessError
import sys
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import cast, TextIO

import toml
import pymysql

from freva_deployment import __version__
from ..utils import (
    logger,
    read_db_credentials,
    download_server_map,
    get_setup_for_service,
    set_log_level,
    guess_map_server,
)

DUMP_SCRIPT = """#!{python_bin}
import json
import sys
sys.path.insert(0, "{python_path}")
from evaluation_system.model.file import DRSFile

with open("{json_file}", "w") as f_obj:
    json.dump(DRSFile.DRS_STRUCTURE, f_obj)
"""


def exec_command(
    command: str, stdout: TextIO | None = None, shell: bool = False
) -> None:
    """Execute a sub process command."""
    logger.debug("Executing command: %s", command)
    env = os.environ.copy()
    _ = env.pop("EVALUATION_SYSTEM_CONFIG_FILE", "")
    std = stdout or PIPE
    res = run(
        shlex.split(command), stdout=std, stderr=PIPE, shell=shell, env=env, check=False
    )
    try:
        res.check_returncode()
    except CalledProcessError as error:
        logger.error("Script returned non-zero exit status:\n%s", command)
        logger.error("STDERROR:\n%s", res.stderr.decode())
        raise error


def execute_script_and_get_config(python_bin: Path, template: str) -> str:
    """Execute a python2 template and read it's ouput."""
    logger.debug("Working on %s", python_bin)
    python_path = python_bin.parents[3] / "freva" / "src"
    with TemporaryDirectory() as temp_dir:
        logger.debug("Creating temporary script in %s", temp_dir)
        with open(Path(temp_dir) / "write.py", "w") as f_obj:
            script = template.format(
                python_bin=python_bin,
                python_path=python_path,
                json_file=Path(temp_dir) / "config.json",
            )
            logger.debug("Writing script:\n%s", script)
            f_obj.write(script)
        exec_command(f"{python_bin} {Path(temp_dir) / 'write.py'}")
        logger.debug("Reading json file:")
        with open(Path(temp_dir) / "config.json") as f_obj:
            return f_obj.read()


def _get_python_path_from_env() -> Path:
    python_path = cast(str, shutil.which("python"))
    return Path(python_path)


def _add_new_db(db_config: dict[str, str], dump_file: str) -> None:
    dump_cmd = (
        f"mysql -h {db_config['db.host']} -u {db_config['db.user']} "
        f"-p'{db_config['db.passwd']}' -P{db_config['db.port']} < {dump_file}"
    )
    logger.debug("Executing command: %s", dump_cmd)
    res = os.system(dump_cmd)
    if res != 0:
        logger.error("Command failed: %s", dump_cmd)
        return
    with pymysql.connect(
        autocommit=True,
        host=db_config["db.host"],
        user=db_config["db.user"],
        password=db_config["db.passwd"],
        port=int(db_config["db.port"]),
        db=db_config["db.db"],
    ) as con:
        with con.cursor() as cursor:
            cursor.execute(
                "SELECT COUNT(*) FROM INFORMATION_SCHEMA.COLUMNS "
                f"WHERE TABLE_SCHEMA='{db_config['db.db']}' AND "
                "TABLE_NAME='history_history' AND column_name='host'"
            )
            results_found = cursor.fetchall()[0][0]
            if results_found == 0:
                logger.info("Adding new column `host` to table.")
                cursor.execute("ALTER TABLE history_history ADD host longtext")


def _migrate_db(parser: argparse.Namespace) -> None:
    db_host = parser.new_hostname
    mysqldump = shutil.which("mysqldump")
    if mysqldump is None:
        logger.error("myslqdump not found, to continue isntall mysqldump")
        return
    new_db_cfg = read_db_credentials(parser.cert_file, db_host)
    with NamedTemporaryFile(suffix=".sql") as temp_file:
        with open(temp_file.name, "w") as f_obj:
            f_obj.write(f"USE `{new_db_cfg['db.db']}`;\n")
            f_obj.flush()
            dump_command = (
                f"{mysqldump} -u {parser.old_user} "
                f"-h {parser.old_hostname} -p'{parser.old_pw}' "
                f"-P{parser.old_port} "
                "--tz-utc --no-create-db "
                f"{parser.old_db}"
            )
            exec_command(dump_command, stdout=f_obj)
            _add_new_db(new_db_cfg, temp_file.name)


def _migrate_drs(parser: argparse.Namespace) -> None:
    python_path = parser.python_path or _get_python_path_from_env()
    config = json.loads(execute_script_and_get_config(python_path, DUMP_SCRIPT))
    logger.debug("Writing toml file to drs_config.toml")
    this_file = (Path(".") / "drs_config.toml").absolute()
    with open(this_file, "w") as f_obj:
        toml.dump(config, f_obj)
    logger.info(
        "Toml file has been written to %s, you can copy the file to its new "
        "location next to the evaluation_system.conf file\n Note: The "
        "evaluation_system.conf location is set by the "
        "EVALUATION_SYSTEM_CONFIG_FILE environment variable after the new "
        "module has been loaded | is sourced.",
        this_file,
    )


def parse_args(argv: list[str]) -> argparse.Namespace:
    """Construct command line argument parser."""
    parser = argparse.ArgumentParser(
        prog="freva-migrate",
        description="Utilities to handle migrations from old freva systems.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    subparsers = parser.add_subparsers(help="Migration commands:", required=True)
    db_parser = subparsers.add_parser(
        "database",
        description="Freva database migration",
        help="Use this command to migrate an existing freva database "
        "to a recently set up system.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    db_parser.add_argument(
        "new_hostname",
        metavar="new_hostname",
        type=str,
        help="The hostname where the new database is deployed.",
    )
    db_parser.add_argument(
        "old_hostname",
        metavar="old_hostname",
        type=str,
        help="Hostname of the old database.",
    )
    db_parser.add_argument(
        "cert_file",
        metavar="cert-file",
        type=Path,
        help="Path to the public certificate file.",
    )
    db_parser.add_argument(
        "--old-port",
        type=int,
        default=3306,
        help="The port where the old database server is running on.",
    )
    db_parser.add_argument(
        "--old-db",
        type=str,
        default="evaluationsystem",
        help="The name of the old database",
    )
    db_parser.add_argument(
        "--old-pw", type=str, default=None, help="The passowrd to the old database"
    )
    db_parser.add_argument(
        "--old-user",
        type=str,
        default="evaluationsystem",
        help="The old database user",
    )
    db_parser.set_defaults(apply_func=_migrate_db)
    return parser.parse_args(argv)


def cli() -> None:
    """Run the command line interface."""
    arg = parse_args(sys.argv[1:])
    set_log_level(arg.verbose)
    arg.apply_func(arg)



"""Module to run the freva deployment."""
from __future__ import annotations
from getpass import getuser
import os
from pathlib import Path
import random
import shlex
import string
from subprocess import run
import sys
from urllib.parse import urlparse
from tempfile import TemporaryDirectory, mkdtemp
from typing import Any

from numpy import sign
import toml
import yaml

from .utils import (
    asset_dir,
    config_dir,
    get_passwd,
    get_email_credentials,
    logger,
    upload_server_map,
    RichConsole,
)


class DeployFactory:
    """Apply freva deployment and its services.

    Parameters
    ----------
    project_name: str
        The name of the project to distinguish this instance from others.
    steps: list[str], default: ["services", "core", "web"]
        The components that are going to be deployed.
    config_file: os.PathLike, default: None
        Path to any existing deployment configuration file.

    Examples
    --------

    >>> from freva_deployment import DeployFactory as DF
    >>> deploy = DF(steps=["solr"])
    >>> deploy.play(ask_pass=True)
    """

    step_order: tuple[str, ...] = ("vault", "db", "solr", "core", "web")
    _steps_with_cert: tuple[str, ...] = ("db", "vault", "core", "web")

    def __init__(
        self,
        steps: list[str] | None = None,
        config_file: Path | str | None = None,
    ) -> None:

        self._config_keys: list[str] = []
        self.master_pass: str = ""
        self.email_password: str = ""
        self._td: TemporaryDirectory = TemporaryDirectory(prefix="inventory")
        self.inventory_file: Path = Path(self._td.name) / "inventory.yaml"
        self.eval_conf_file: Path = Path(self._td.name) / "evaluation_system.conf"
        self.web_conf_file: Path = Path(self._td.name) / "freva_web.toml"
        self.apache_config: Path = Path(self._td.name) / "freva_web.conf"
        self._db_pass: str = ""
        self._steps = steps or ["services", "core", "web"]
        self._inv_tmpl = Path(config_file or config_dir / "inventory.toml")
        self._cfg_tmpl = self.aux_dir / "evaluation_system.conf.tmpl"
        self.cfg = self._read_cfg()
        self.project_name = self.cfg.pop("project_name", None)
        if not self.project_name:
            raise ValueError("You must set a project name")

    @property
    def public_key_file(self) -> str:
        """Path to the public certificate file."""
        public_keyfile = self.cfg["certificates"].get("public_keyfile")
        chain_keyfile = self.cfg["certificates"].get("chain_keyfile")
        if public_keyfile:
            return str(Path(public_keyfile).expanduser().absolute())
        if chain_keyfile:
            return str(Path(chain_keyfile).expanduser().absolute())
        raise ValueError("You must give a valid path to a public key file.")

    @property
    def private_key_file(self) -> str:
        """Path to the private key file."""
        keyfile = self.cfg["certificates"].get("private_keyfile")
        if keyfile:
            return str(Path(keyfile).expanduser().absolute())
        raise ValueError("You must give a valid path to a private key file.")

    @property
    def chain_key_file(self) -> str:
        """Path to the private key file."""
        keyfile = self.cfg["certificates"].get("chain_keyfile")
        if keyfile:
            return str(Path(keyfile).expanduser().absolute())
        return ""

    def _prep_vault(self) -> None:
        """Prepare the vault."""
        self._config_keys.append("vault")
        self.cfg["vault"] = self.cfg["db"].copy()
        if not self.master_pass:
            self.master_pass = get_passwd()
        self.cfg["vault"]["config"]["root_passwd"] = self.master_pass
        self.cfg["vault"]["config"]["passwd"] = self.db_pass
        self.cfg["vault"]["config"]["keyfile"] = self.public_key_file
        self.cfg["vault"]["config"]["email"] = self.cfg["web"]["config"].get(
            "contacts", ""
        )

    def _prep_db(self) -> None:
        """prepare the mariadb service."""
        self._config_keys.append("db")
        if not self.master_pass:
            self.master_pass = get_passwd()
        host = self.cfg["db"]["hosts"]
        self.cfg["db"]["config"]["root_passwd"] = self.master_pass
        self.cfg["db"]["config"]["passwd"] = self.db_pass
        self.cfg["db"]["config"]["keyfile"] = self.public_key_file
        for key in ("name", "user", "db"):
            self.cfg["db"]["config"][key] = self.cfg["db"]["config"].get(key) or "freva"
        db_host = self.cfg["db"]["config"].get("host", "")
        if not db_host:
            self.cfg["db"]["config"]["host"] = host
        self.cfg["db"]["config"].setdefault("port", "3306")
        self.cfg["db"]["config"]["email"] = self.cfg["web"]["config"].get(
            "contacts", ""
        )
        self._prep_vault()

    def _prep_solr(self) -> None:
        """prepare the apache solr service."""
        self._config_keys.append("solr")
        self.cfg["solr"]["config"].pop("core", None)
        for key, default in dict(mem="4g", port=8983).items():
            self.cfg["solr"]["config"][key] = (
                self.cfg["solr"]["config"].get(key) or default
            )
        self.cfg["solr"]["config"]["email"] = self.cfg["web"]["config"].get(
            "contacts", ""
        )

    def _prep_core(self) -> None:
        """prepare the core deployment."""
        self._config_keys.append("core")
        self.cfg["core"]["config"]["admins"] = (
            self.cfg["core"]["config"].get("admins") or getuser()
        )
        if not self.cfg["core"]["config"]["admins"]:
            self.cfg["core"]["config"]["admins"] = getuser()
        install_dir = self.cfg["core"]["config"]["install_dir"]
        root_dir = self.cfg["core"]["config"].get("root_dir", "")
        if not root_dir:
            self.cfg["core"]["config"]["root_dir"] = install_dir
        preview_path = self.cfg["core"]["config"].get("preview_path", "")
        base_dir_location = self.cfg["core"]["config"].get("base_dir_location", "")
        scheduler_output_dir = self.cfg["core"]["config"].get(
            "scheduler_output_dir", ""
        )
        scheduler_system = self.cfg["core"]["config"].get("scheduler_system", "local")
        if not preview_path:
            if base_dir_location:
                self.cfg["core"]["config"]["preview_path"] = str(
                    Path(base_dir_location) / "share" / "preview"
                )
            else:
                self.cfg["core"]["config"]["preview_path"] = ""
        if not scheduler_output_dir:
            scheduler_output_dir = str(Path(base_dir_location) / "share")
        scheduler_output_dir = Path(scheduler_output_dir) / scheduler_system
        self.cfg["core"]["config"]["scheduler_output_dir"] = str(scheduler_output_dir)
        self.cfg["core"]["config"]["keyfile"] = self.public_key_file
        git_exe = self.cfg["core"]["config"].get("git_path")
        self.cfg["core"]["config"]["git_path"] = git_exe or "git"
        self.cfg["core"]["config"][
            "git_url"
        ] = "https://gitlab.dkrz.de/freva/evaluation_system.git"

    def _prep_web(self) -> None:
        """prepare the web deployment."""
        self._config_keys.append("web")
        self._prep_core()
        admin = self.cfg["core"]["config"]["admins"]
        if not isinstance(admin, str):
            self.cfg["web"]["config"]["admin"] = admin[0]
        else:
            self.cfg["web"]["config"]["admin"] = admin
        _webserver_items = {}
        try:
            _webserver_items = {
                k.replace("web_", "").upper(): v
                for (k, v) in self.cfg["web"]["config"].items()
            }
        except KeyError as error:
            raise KeyError(
                "No web config section given, please configure the web.config"
            ) from error
        _webserver_items["ALLOWED_HOSTS"].append(self.cfg["web"]["hosts"])
        _webserver_items["REDIS_HOST"] = self.cfg["web"]["hosts"]
        try:
            with Path(_webserver_items["homepage_text"]).open("r") as f_obj:
                _webserver_items["homepage_text"] = f_obj.read()
        except (FileNotFoundError, IOError, KeyError):
            pass
        server_name = self.cfg["web"]["config"].pop("server_name", [])
        if isinstance(server_name, str):
            server_name = server_name.split(",")
        server_name = ",".join([a for a in server_name if a.strip()])
        if not server_name:
            server_name = self.cfg["web"]["hosts"]
        self.cfg["web"]["config"]["server_name"] = server_name
        web_host = self.cfg["web"]["hosts"]
        if web_host == "127.0.0.1":
            web_host = "localhost"
        self.cfg["web"]["config"]["host"] = web_host
        _webserver_items["CSRF_TRUSTED_ORIGINS"] = []
        for url in (server_name, self.cfg["web"]["config"]["project_website"]):
            trusted_origin = urlparse(url)

            if trusted_origin.scheme:
                _webserver_items["CSRF_TRUSTED_ORIGINS"].append(
                    f"https://{trusted_origin.netloc}"
                )
            else:
                _webserver_items["CSRF_TRUSTED_ORIGINS"].append(
                    f"https://{trusted_origin.path}"
                )
        _webserver_items["FREVA_BIN"] = os.path.join(
            self.cfg["core"]["config"]["install_dir"], "bin"
        )
        try:
            with Path(_webserver_items["ABOUT_US_TEXT"]).open() as f_obj:
                _webserver_items["ABOUT_US_TEXT"] = f_obj.read()
        except (FileNotFoundError, IOError, KeyError):
            pass
        try:
            _webserver_items["IMPRINT"] = _webserver_items["IMPRINT"].split(",")
        except AttributeError:
            pass
        with self.web_conf_file.open("w") as f_obj:
            toml.dump(_webserver_items, f_obj)
        for key in ("core", "web"):
            self.cfg[key]["config"]["config_toml_file"] = str(self.web_conf_file)
        if not self.master_pass:
            self.master_pass = get_passwd()
        email_user, self.email_password = get_email_credentials()
        self._prep_vault()
        self.cfg["vault"]["config"]["ansible_python_interpreter"] = self.cfg["db"][
            "config"
        ].get("ansible_python_interpreter", "/usr/bin/python3")
        self.cfg["vault"]["config"]["email_user"] = email_user
        self.cfg["vault"]["config"]["email_password"] = self.email_password
        self.cfg["web"]["config"]["root_passwd"] = self.master_pass
        self.cfg["web"]["config"]["private_keyfile"] = self.private_key_file
        self.cfg["web"]["config"]["public_keyfile"] = self.public_key_file
        self.cfg["web"]["config"]["chain_keyfile"] = (
            self.chain_key_file or self.public_key_file
        )
        self.cfg["web"]["config"]["apache_config_file"] = str(self.apache_config)
        self._prep_apache_config()

    def _prep_apache_config(self):
        config = []
        with (Path(asset_dir) / "web" / "freva_web.conf").open() as f_obj:
            for line in f_obj.readlines():
                if not self.chain_key_file and "SSLCertificateChainFile" in line:
                    continue
                config.append(line)
        with open(self.apache_config, "w") as f_obj:
            f_obj.write("".join(config))

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._td.cleanup()

    def _read_cfg(self) -> dict[str, Any]:
        try:
            with self._inv_tmpl.open() as f_obj:
                return dict(toml.load(f_obj))
        except FileNotFoundError as error:
            raise FileNotFoundError(f"No such file {self._inv_tmpl}") from error

    def _check_config(self) -> None:
        sections = []
        for section in self.cfg.keys():
            for step in self._config_keys:
                if section.startswith(step) and section not in sections:
                    sections.append(section)
        for section in sections:
            for key, value in self.cfg[section]["config"].items():
                if not value and not self._empty_ok and not isinstance(value, bool):
                    raise ValueError(f"{key} in {section} is empty in {self._inv_tmpl}")

    @property
    def _empty_ok(self) -> list[str]:
        """Define all keys that can be empty."""
        return [
            "admins",
            "conda_exec_path",
        ]

    def _get_files_copy(self, key) -> Path | None:
        return dict(
            core=self.eval_conf_file.absolute(),
            web=self.eval_conf_file.absolute(),
        ).get(key, None)

    @property
    def db_pass(self) -> str:
        """Create a password for the database."""
        if self._db_pass:
            return self._db_pass
        punctuations = "!@$^&*()_+-;:|,.%"
        num_chars, num_digits, num_punctuations = 20, 4, 4
        num_chars -= num_digits + num_punctuations
        characters = [
            "".join([random.choice(string.ascii_letters) for i in range(num_chars)]),
            "".join([random.choice(string.digits) for i in range(num_digits)]),
            "".join([random.choice(punctuations) for i in range(num_punctuations)]),
        ]
        str_characters = "".join(characters)
        _db_pass = "".join(random.sample(str_characters, len(str_characters)))
        while _db_pass.startswith("@"):
            # Vault treats values starting with "@" as file names.
            _db_pass = "".join(random.sample(str_characters, len(str_characters)))
        self._db_pass = _db_pass
        return self._db_pass

    @property
    def _needs_core(self) -> list[str]:
        """Define the steps that need the core config."""
        return ["web", "core"]

    def _set_additional_config_values(
        self, step: str, config: dict[str, dict[str, dict[str, str | int | bool]]]
    ) -> None:
        """Set additional values to the configuration."""
        if step in self._needs_core:
            for key in (
                "root_dir",
                "base_dir_location",
                "preview_path",
                "scheduler_output_dir",
            ):
                value = self.cfg["core"]["config"].get(key, "")
                config[step]["vars"][f"core_{key}"] = value
        config[step]["vars"][f"{step}_hostname"] = self.cfg[step]["hosts"]
        config[step]["vars"][f"{step}_name"] = f"{self.project_name}-{step}"
        config[step]["vars"]["asset_dir"] = str(asset_dir)
        config[step]["vars"]["ansible_user"] = (
            self.cfg[step]["config"].get("ansible_user") or getuser()
        )
        config[step]["vars"][f"{step}_ansible_python_interpreter"] = (
            self.cfg[step]["config"].get("ansible_python_interpreter")
            or "/usr/bin/python3"
        )
        dump_file = self._get_files_copy(step)
        if dump_file:
            config[step]["vars"][f"{step}_dump"] = str(dump_file)

    def parse_config(self) -> str:
        """Create config files for anisble and evaluation_system.conf."""
        logger.info("Parsing configurations")
        self._check_config()
        config: dict[str, dict[str, dict[str, str | int | bool]]] = {}
        for step in set(self._config_keys):
            config[step] = {}
            config[step]["hosts"] = self.cfg[step]["hosts"]
            config[step]["vars"] = {}
            for key, value in self.cfg[step]["config"].items():
                if key in ("root_passwd", "wipe"):
                    new_key = key
                else:
                    new_key = f"{step}_{key}"
                config[step]["vars"][new_key] = value
            config[step]["vars"]["project_name"] = self.project_name
            # Add additional keys
            self._set_additional_config_values(step, config)
        return yaml.dump(config)

    @property
    def _playbook_file(self) -> Path:

        return Path(self._td.name) / "ansible-playbook.yml"

    @property
    def python_prefix(self) -> Path:
        """Get the path of the new conda evnironment."""
        return Path(sys.exec_prefix) / "bin" / "python3"

    @property
    def aux_dir(self) -> Path:
        """Directory with auxillary files."""
        return asset_dir / "config"

    @property
    def playbook_dir(self) -> Path:
        """The location of all playbooks."""
        return asset_dir / "playbooks"

    @property
    def steps(self) -> list[str]:
        """Set all the deploment steps."""
        steps = []
        for step in self._steps:
            steps.append(step)
            if step == "db":
                steps.append("vault")
        return [s for s in self.step_order if s in steps]

    def create_playbooks(self):
        """Create the ansible playbook form all steps."""
        logger.info("Creating Ansible playbooks")
        playbook = []
        for step in self.steps:
            getattr(self, f"_prep_{step}")()
            playbook_file = self.playbook_dir / f"{step}-server-playbook.yml"
            with playbook_file.open() as f_obj:
                playbook += yaml.safe_load(f_obj)
        with self._playbook_file.open("w") as f_obj:
            yaml.dump(playbook, f_obj)

    def create_eval_config(self) -> None:
        """Create and dump the evaluation_systme.config."""
        logger.info("Creating evaluation_system.conf")
        keys = (
            ("core", "admins"),
            ("web", "project_website"),
            ("core", "root_dir"),
            ("core", "base_dir_location"),
            ("core", "preview_path"),
            ("core", "scheduler_output_dir"),
            ("core", "scheduler_system"),
        )
        cfg_file = asset_dir / "config" / "evaluation_system.conf.tmpl"
        with cfg_file.open() as f_obj:
            lines = f_obj.readlines()
            for num, line in enumerate(lines):
                if line.startswith("project_name"):
                    lines[num] = f"project_name={self.project_name}\n"
                for key, value in keys:
                    if line.startswith(value):
                        cfg = self.cfg[key]["config"].get(value, "")
                        if cfg:
                            lines[num] = f"{value}={cfg}\n"
                for step in ("solr", "db"):
                    cfg = self.cfg[step]["config"].get("port", "")
                    if line.startswith(f"{step}.port"):
                        lines[num] = f"{step}.port={cfg}\n"
                    if line.startswith(f"{step}.host"):
                        lines[num] = f"{step}.host={self.cfg[step]['hosts']}\n"
        dump_file = self._get_files_copy("core")
        if dump_file:
            with dump_file.open("w") as f_obj:
                f_obj.write("".join(lines))

    def play(
        self, server_map: str | None = None, ask_pass: bool = True, verbosity: int = 0
    ) -> None:
        """Play the ansible playbook.

        Parameters
        ----------
        server_map: str, default: None
            Hostname running the service that keeps track of the server
            infrastructure, if None given (default) no new deployed services
            are added.
        ask_pass: bool, default: True
            Instruct Ansible to ask for the ssh passord instead of using a
            ssh key
        verbosity: int, default: 0
            Verbosity level, default 0
        """
        self.create_playbooks()
        inventory = self.parse_config()
        self.create_eval_config()
        with self.inventory_file.open("w") as f_obj:
            f_obj.write(inventory)
        inventory_str = inventory
        for passwd in (self.email_password, self.master_pass):
            if passwd:
                inventory_str = inventory_str.replace(passwd, "*" * len(passwd))
        RichConsole.print(inventory, style="bold", markup=True)
        logger.info("Playing the playbooks with ansible")
        RichConsole.print(
            "[b]Note:[/] The [blue]BECOME[/] password refers to the "
            "[blue]sudo[/] password",
            markup=True,
        )
        v_string = sign(verbosity) * "-" + verbosity * "v"
        cmd = (
            f"ansible-playbook {v_string} -i {self.inventory_file} "
            f"{self._playbook_file} --ask-become-pass"
        )
        if ask_pass:
            cmd += " --ask-pass"
        try:
            _ = run(
                shlex.split(cmd), env=os.environ.copy(), cwd=str(asset_dir), check=True
            )
        except KeyboardInterrupt:
            pass
        if server_map:
            self.upload_server_info(server_map, inventory)

    def upload_server_info(self, server_map: str, inventory: str) -> None:
        """Upload information on server information to shared nextcloud."""
        deployment_setup: dict[str, dict[str, str]] = {}
        for step, info in yaml.safe_load(inventory).items():
            ansible_step = dict(
                python_path=info["vars"][f"{step}_ansible_python_interpreter"]
            )
            ansible_step["hosts"] = info["hosts"]
            deployment_setup[step] = ansible_step
            if step == "web":
                deployment_setup["redis"] = ansible_step
        upload_server_map(server_map, self.project_name, deployment_setup)



"""Command line interface for the deployment."""
from __future__ import annotations
import argparse
from pathlib import Path
import sys

from freva_deployment import __version__
from ..deploy import DeployFactory
from ..utils import config_dir, set_log_level, guess_map_server


def parse_args(argv: list[str] | None) -> argparse.Namespace:
    """Consturct command line argument parser."""

    ap = argparse.ArgumentParser(
        prog="deploy-freva-cmd",
        description="Deploy freva and its services on different machines.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "--server-map",
        type=str,
        default=None,
        help=(
            "Hostname of the service mapping the freva server "
            "archtiecture, Note: you can create a server map by "
            "running the deploy-freva-map command"
        ),
    )
    ap.add_argument(
        "--config",
        "-c",
        type=Path,
        help="Path to ansible inventory file.",
        default=config_dir / "inventory.toml",
    )
    ap.add_argument(
        "--steps",
        type=str,
        nargs="+",
        default=["db", "solr", "web", "core"],
        choices=["web", "core", "db", "solr"],
        help="The services/code stack to be deployed",
    )
    ap.add_argument(
        "--ask-pass",
        help="Connect to server via ssh passwd instead of public key.",
        action="store_true",
        default=False,
    )
    ap.add_argument(
        "-v", "--verbose", action="count", help="Verbosity level", default=0
    )
    ap.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    return ap.parse_args()


def cli(argv: list[str] | None = None) -> None:
    """Run the command line interface."""
    args = parse_args(argv)
    if args.server_map:
        server_map = guess_map_server(args.server_map, mandatory=False)
    else:
        server_map = ""
    set_log_level(args.verbose)
    with DeployFactory(
        steps=args.steps,
        config_file=args.config,
    ) as DF:
        DF.play(server_map, args.ask_pass, args.verbose)


if __name__ == "__main__":
    cli(sys.argv[1:])



"""The Freva deployment Text User Interface (TUI) helps to configure a
deployment setup for a new instance of freva."""
from __future__ import annotations
import json
from pathlib import Path
import time
import threading
from typing import Any, Dict, List, cast

import appdirs
import npyscreen
import tomlkit

from freva_deployment.utils import asset_dir, config_dir
from .base import selectFile, BaseForm
from .deploy_forms import WebScreen, DBScreen, SolrScreen, CoreScreen, RunForm


class MainApp(npyscreen.NPSAppManaged):
    config: dict[str, Any] = dict()

    @property
    def steps(self) -> list[str]:
        """Get the deploy steps."""
        steps = []
        for step, form_obj in self._forms.items():
            if form_obj.use.value and step not in steps:
                steps.append(step)
        return steps

    def onStart(self) -> None:
        """When Application starts, set up the Forms that will be used."""
        self.cache_dir.mkdir(exist_ok=True, parents=True)
        self.setup: dict[str, Any] = {}
        self._steps_lookup = {
            "core": "MAIN",
            "web": "SECOND",
            "solr": "FOURTH",
            "db": "THIRD",
        }
        self._forms: dict[str, BaseForm] = {}
        self.current_form = "core"
        self.config = cast(Dict[str, Any], self._read_cache("config", {}))
        for step in self._steps_lookup.keys():
            self.config.setdefault(step, {"hosts": "", "config": {}})
        self._add_froms()
        self.start_auto_save()

    def start_auto_save(self) -> None:
        """(Re)-Start the auto save thread."""
        self.thread_stop = threading.Event()
        self._save_thread = threading.Thread(target=self._auto_save)
        self._save_thread.start()

    def _add_froms(self) -> None:
        """Add forms to edit the deploy steps to the main window."""
        self._forms["core"] = self.addForm(
            "MAIN",
            CoreScreen,
            name="Core deployment",
        )
        self._forms["web"] = self.addForm("SECOND", WebScreen, name="Web deployment")
        self._forms["db"] = self.addForm("THIRD", DBScreen, name="Database deployment")
        self._forms["solr"] = self.addForm("FOURTH", SolrScreen, name="Solr deployment")
        self._setup_form = self.addForm("SETUP", RunForm, name="Apply the Deployment")

    def exit_application(self, *args, **kwargs) -> None:
        value = npyscreen.notify_ok_cancel(
            kwargs.get("msg", "Exit Application?"), title=""
        )
        if value is True:
            self.thread_stop.set()
            self.setNextForm(None)
            self.save_config_to_file()
            self.editing = False
            self.switchFormNow()

    def change_form(self, name: str) -> None:
        # Switch forms.  NB. Do *not* call the .edit() method directly (which
        # would lead to a memory leak and ultimately a recursion error).
        # Instead, use the method .switchForm to change forms.
        self.switchForm(name)

        # By default the application keeps track of every form visited.
        # There's no harm in this, but we don't need it so:
        self.resetHistory()

    def check_missing_config(self, stop_at_missing: bool = True) -> str | None:
        """Evaluate all forms."""
        for step, form_obj in self._forms.items():
            cfg = form_obj.check_config(notify=stop_at_missing)
            if cfg is None and stop_at_missing:
                return self._steps_lookup[step]
            self.config[step] = cfg
        return None

    def _auto_save(self) -> None:
        """Auto save the current configuration."""
        while not self.thread_stop.is_set():
            time.sleep(0.5)
            if self.thread_stop.is_set():
                break
            try:
                self.check_missing_config(stop_at_missing=False)
                self.save_config_to_file()
            except Exception:
                pass

    def save_dialog(self, *args, **kwargs) -> None:
        """Create a dialoge that allows for saving the config file."""

        the_selected_file = selectFile(
            select_dir=False, must_exist=False, file_extentions=[".toml"]
        )
        if the_selected_file:
            the_selected_file = Path(the_selected_file).with_suffix(".toml")
            the_selected_file = str(the_selected_file.expanduser().absolute())
            self.check_missing_config(stop_at_missing=False)
            self._setup_form.inventory_file.value = the_selected_file
            self.save_config_to_file(write_toml_file=True)

    def _update_config(self, config_file: Path | str) -> None:
        """Update the main window after a new configuration has been loaded."""

        try:
            with open(config_file) as f:
                self.config = tomlkit.load(f)
        except Exception:
            return
        self.resetHistory()
        self.editing = True
        self.switchFormNow()
        self._add_froms()
        self._setup_form.inventory_file.value = config_file

    def load_dialog(self, *args, **kwargs) -> None:
        """Create a dialoge that allows for loading a config file."""

        the_selected_file = selectFile(
            select_dir=False, must_exist=True, file_extentions=[".toml"]
        )
        if the_selected_file:
            self._update_config(the_selected_file)
            self.save_config_to_file()

    def save_config_to_file(self, **kwargs) -> Path | None:
        """Save the status of the tui to file."""
        try:
            return self._save_config_to_file(**kwargs)
        except Exception as error:
            npyscreen.notify_confirm(
                title="Error",
                message=f"Couldn't save config:\n{error}",
            )
        return None

    def _save_config_to_file(self, write_toml_file: bool = False) -> Path | None:
        cache_file = self.cache_dir / ".temp_file.toml"
        save_file = self._setup_form.inventory_file.value
        if save_file:
            save_file = str(Path(save_file).expanduser().absolute())
        else:
            save_file = None
        cert_files = dict(
            public_keyfile=self._setup_form.public_keyfile.value or "",
            private_keyfile=self._setup_form.private_keyfile.value or "",
            chain_keyfile=self._setup_form.chain_keyfile.value or "",
        )
        for key, value in cert_files.items():
            if value:
                cert_files[key] = str(Path(value).expanduser().absolute())
        project_name = self._setup_form.project_name.value
        server_map = self._setup_form.server_map.value
        ssh_pw = self._setup_form.use_ssh_pw.value
        if isinstance(ssh_pw, list):
            ssh_pw = bool(ssh_pw[0])
        config = {
            "save_file": save_file,
            "steps": self.steps,
            "project_name": project_name,
            "ssh_pw": ssh_pw,
            "server_map": server_map,
            "config": self.config,
        }
        config.update(cert_files)
        with open(self.cache_dir / "freva_deployment.json", "w") as f:
            json.dump({k: v for (k, v) in config.items()}, f, indent=3)
        if write_toml_file is False:
            return None
        save_file = Path(save_file or cache_file)
        try:
            with open(asset_dir / "config" / "inventory.toml") as f:
                config_tmpl = cast(Dict[str, Any], tomlkit.load(f))
        except Exception:
            config_tmpl = self.config
        config_tmpl["certificates"] = cert_files
        config_tmpl["project_name"] = project_name
        for step, settings in self.config.items():
            if step in ("certificates", "project_name"):
                continue
            config_tmpl[step]["hosts"] = settings["hosts"]
            for key, config in settings["config"].items():
                config_tmpl[step]["config"][key] = config
        save_file.parent.mkdir(exist_ok=True, parents=True)
        with open(save_file, "w") as f:
            toml_string = tomlkit.dumps(config_tmpl)
            f.write(toml_string)
        return save_file

    @property
    def cache_dir(self) -> Path:
        """The user cachedir."""
        return Path(appdirs.user_cache_dir()) / "freva-deployment"

    def _read_cache(
        self, key: str, default: str | list | bool | dict[str, str] = ""
    ) -> str | bool | list | dict[str, str]:
        try:
            with open(self.cache_dir / "freva_deployment.json", "r") as f:
                return json.load(f).get(key, default)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            return default

    @property
    def _steps(self) -> list[str]:
        """Read the deployment-steps from the cache."""
        return cast(List[str], self._read_cache("steps", ["core", "web", "db", "solr"]))

    def read_cert_file(self, key: str) -> str:
        """Read the certificate file from the cache."""
        cert_file = cast(str, self.config.get("certificates", {}).get(key, ""))
        if cert_file:
            return cert_file
        return cast(str, self._read_cache(key, ""))

    @property
    def save_file(self) -> str:
        """Read the file that stores the configuration from the cache."""
        fall_back = config_dir / "inventory.toml"
        return cast(str, self._read_cache("save_file", str(fall_back)))


if __name__ == "__main__":
    try:
        main_app = MainApp()
        main_app.run()
    except KeyboardInterrupt:
        pass
